/* DS1307 RTC Basic Demo
 * http://frank.circleofcurrent.com/musicclock/
 * Copyright (c) 2011 Frank Zhao
 *
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

// please see http://www.nongnu.org/avr-libc/user-manual/modules.html for info on these modules

#include <avr/io.h> // allows access to AVR built-in peripherals and registers
#include <avr/pgmspace.h> // allows access to flash memory
#include <util/delay.h> // provides delays
#include <avr/interrupt.h> // we need sei() from here to enable interrupts
#include <stdio.h> // allows formatted debug output
#include <avr/power.h> // used for setting clock prescaler
#include "libs/usb_serial.h" // provides USB-to-serial functionality
#include "libs/lcd.h" // provides LCD display functionality

// binary coded decimal macros
#define bcd_encode(x) ((((x)/10)<<4)+((x)%10))
#define bcd_decode(x) (((x)&0x0F)+((((x)&0xF0)>>4)*10))

// pin definitions
#define TWI_PORTDIR DDRD
#define TWI_PORTOUTPUT PORTD
#define TWI_PORTINPUT PIND
#define TWI_PIN_SDA 1
#define TWI_PIN_SCL 0

#define DS1307_ADDR 0b01101000 // TWI address of the DS1307


// this is a wrapper for usb_putchar that allows me to use printf from stdio.h
static inline int debug_putChar(char c, FILE* stream)
{
    return usb_serial_putchar(c); // refer to http://www.pjrc.com/teensy/usb_serial.html
}

// this is a wrapper for lcd_sendchar that allows me to use printf from stdio.h
static inline int lcd_putChar(char c, FILE* stream)
{
    lcd_sendChar(c);
}

// setup debug stream for printf from stdio.h
static FILE debug_stdout = FDEV_SETUP_STREAM(debug_putChar, NULL, _FDEV_SETUP_WRITE);

// setup stream for printf from stdio.h using the LCD display
static FILE lcd_stdout = FDEV_SETUP_STREAM(lcd_putChar, NULL, _FDEV_SETUP_WRITE);

void rtc_init()
{
    // port must be set as input but with internal pull-up resistors enabled
    TWI_PORTOUTPUT |= _BV(TWI_PIN_SDA);
    TWI_PORTOUTPUT |= _BV(TWI_PIN_SCL);
    TWI_PORTDIR &= ~_BV(TWI_PIN_SDA);
    TWI_PORTDIR &= ~_BV(TWI_PIN_SCL);
    
    // start TWI
    twi_init();
}

int main()
{
    clock_prescale_set(clock_div_2); // this is because we are running at 3.3V
    // see http://www.nongnu.org/avr-libc/user-manual/group__avr__power.html and http://www.pjrc.com/teensy/3volt.html
    
    DDRD |= _BV(6);
    PORTD &= ~_BV(6);
    
    usb_init(); // initialize usb obviously
    
    while (usb_configured() == 0) _delay_ms(5); // waits until USB is ready
    
    // 5 seconds for you to connect hyperterminal
    // while flashing backlight
    LCD_DDR |= _BV(LCD_PIN_LED);
    for (int i = 0; i < 5; i++)
    {
        lcd_backlightOn();
        _delay_ms(500);
        lcd_backlightOff();
        _delay_ms(500);
    }
    
    PORTD |= _BV(6);
    
    // here's the start of the demonstration
    fprintf(&debug_stdout, "Initializing LCD\r\n");
    lcd_init();
    lcd_backlightOn();
    
    fprintf(&debug_stdout, "Initializing TWI for RTC\r\n");
    rtc_init();
    sei();
    
    uint8_t buffer[7];
    uint8_t errCode;
    
    fprintf(&debug_stdout, "Perform Single Byte Write to DS1307 RTC\r\n");
    
    buffer[0] = 0x00; // set ds1307 internal register address pointer
    buffer[1] = 12; // the data to set the internal register to
    errCode = twi_writeTo(DS1307_ADDR, buffer, 2);
    fprintf(&debug_stdout, "Result Code: %d\r\n", errCode);
    
    fprintf(&debug_stdout, "Perform Multi Byte Write to DS1307 RTC (sending 1 2 3 4)\r\n");
    
    buffer[0] = 0x01; // set ds1307 internal register address pointer
    buffer[1] = 1; // the data to set the internal register 0x01 to
    buffer[2] = 2; // the data to set the internal register 0x02 to
    buffer[3] = 3; // the data to set the internal register 0x03 to
    buffer[4] = 4; // the data to set the internal register 0x04 to
    errCode = twi_writeTo(DS1307_ADDR, buffer, 5);
    fprintf(&debug_stdout, "Result Code: %d\r\n", errCode);
    
    fprintf(&debug_stdout, "Perform Single Byte Read from DS1307 RTC\r\n");
    buffer[0] = 0x01; // set ds1307 internal register address pointer
    errCode = twi_writeTo(DS1307_ADDR, buffer, 1);
    fprintf(&debug_stdout, "Result Code for twi_writeTo: %d\r\n", errCode);
    twi_readFrom(DS1307_ADDR, buffer, 1);
    fprintf(&debug_stdout, "DS1307 Replied: %d\r\n", buffer[0]);
    
    fprintf(&debug_stdout, "Perform Multi Byte Read from DS1307 RTC\r\n");
    buffer[0] = 0x01; // set ds1307 internal register address pointer
    errCode = twi_writeTo(DS1307_ADDR, buffer, 1);
    fprintf(&debug_stdout, "Result Code for twi_writeTo: %d\r\n", errCode);
    twi_readFrom(DS1307_ADDR, buffer, 4);
    fprintf(&debug_stdout, "DS1307 Replied: %d, %d, ", buffer[0], buffer[1]);
    fprintf(&debug_stdout, "%d, %d\r\n", buffer[2], buffer[3]);
    
    fprintf(&debug_stdout, "Entering Time Display Demo (MM:SS on LCD)\r\n");
    
    PORTD &= ~_BV(6);
    
    while (1)
    {
        uint8_t minute;
        buffer[0] = 0x01; // set ds1307 internal register address
        twi_writeTo(DS1307_ADDR, buffer, 1);
        twi_readFrom(DS1307_ADDR, buffer, 1);
        minute = bcd_decode(buffer[0]);
        
        uint8_t second;
        buffer[0] = 0x00; // set ds1307 internal register address
        twi_writeTo(DS1307_ADDR, buffer, 1);
        twi_readFrom(DS1307_ADDR, buffer, 1);
        second = bcd_decode(buffer[0]);
        
        fprintf(&debug_stdout, "%d:%d\r\n", minute, second);
        
        lcd_printTime(minute, second, 0);
        
        _delay_ms(250);
    }
    
    return 0;
}